"""
The status command is used to determine the whether any or all of the 4SS
servers are running.  For a listing all of the servers, execute the status
command with out any arguments.  To get the status of an individual server,
specify the name(s) on the command line.
"""

import os, sys

from Ft.Server.Server import FtServerServerException, Error
from Ft.Server.Core import Constants
from Ft.Server.Core.Lib import ConfigFile


def Run(serverNames=None, configFile=None, coreId=None):
    from Ft.Server.Protocols import Launcher, StartMode
    properties = ConfigFile.ReadProperties(fileName=configFile)
    coreId = coreId or Constants.DEFAULT_CORE_ID

    serverDefs = {}
    for defs in properties.values():
        serverDefs.update(defs)

    allServers = 0
    if not serverNames:
        serverNames = []
        allServers = 1
        for name, prop in serverDefs.items():
            if prop.has_key('StartMode'):
                # It is a server entry
                try:
                    launcher = Launcher.Launcher(name, prop, display=1,configFileName=configFile,coreId=coreId)
                except FtServerServerException, exc:
                    if exc.code != Error.NOT_SUPPORTED:
                        raise
                serverNames.append(name)

    for name in serverNames:
        prop = serverDefs.get(name)
        if not prop:
            raise FtServerServerException(Error.UNKNOWN_SERVER, server=name)
        launcher = Launcher.Launcher(name, prop, 1,configFileName=configFile,coreId=coreId)

        (pid, status) = launcher.status()
        if status == Launcher.PID_RUNNING:
            # anything that is running, restart it
            launcher.restart()
        elif prop['StartMode'] == StartMode.AUTOMATIC or not allServers:
            # not running, but should be, start it
            launcher.start()


    return


def TextRestart(options, args):
    try:
        serverNames = args['serverName']
        configFile = options.get('config-file')
        coreId = options.get('core-id')
        Run(serverNames, configFile, coreId)
    except FtServerServerException, e:
        print e
    return


def Register():
    from Ft.Lib.CommandLine import Options, Command, Arguments

    cmd = Command.Command('restart',
                          "Restart server(s)",
                          'OmniOrbServer',
                          __doc__,
                          function=TextRestart,
                          arguments = [Arguments.ZeroOrMoreArgument('serverName',
                                                                    'Name of server to wuery',
                                                                    str),
                                       ],
                          authenticate = 'start',
                          )
    return cmd

