########################################################################
#
# File Name:            Set.SysAcl.py
#
# Documentation:        http://docs.4suite.org/4SS/Lib/CommandLine.py.html
#
"""
WWW: http://4suite.org        e-mail: support@4suite.org

Copyright (c) 2001 Fourthought Inc, USA.   All Rights Reserved.
See  http://4suite.org/COPYRIGHT  for license and copyright information
"""

__doc__ = """Set the access control information the system object.  You need write access to the object to change its ACL values.."""

from Ft.Server.Server import FtServerException
from Ft.Server.Console import CommandUtil, Authenticate
from Ft.Server.Core import Constants
def Run(options,args):

    #Do authentication our selve
    options['AUTHENTICATION'] = Authenticate.UserAuthenticate(options=options)

    try:
        commit = 0
        repo = CommandUtil.GetRepository(options)
        try:

            so = repo.getSystemObject()
            for k in ['read',
                      'write',
                      'init',
                      'start',
                      'stop']:
                ident = options.get(k)
                if ident:
                    so.setAcl(k,ident)
            commit = 1
        finally:
            if commit:
                repo.txCommit()
            else:
                repo.txRollback()
    except FtServerException, e:
        raise


def Register():

    from Ft.Lib.CommandLine import Options, Command, Arguments

    cmd = Command.Command('sysacl',
                          'Set Acl information on the system object.',
                          '--write=u_molson',
                          __doc__,
                          function = Run,
                          arguments = [
                                       ],
                          options = Options.Options([Options.Option(None,'init=','New init permissions'),
                                                    Options.Option(None,'start=','New start permissions'),
                                                    Options.Option(None,'stop=','New start permissions'),
                                                    Options.Option(None,'write=','New Write permissions'),
                                                    Options.Option(None,'read=','New Read permissions'),
                                                    ]),
                          )
    return cmd




