"""
The status command is used to determine the whether any or all of the 4SS
servers are running.  For a listing all of the servers, execute the status
command with out any arguments.  To get the status of an individual server,
specify the name(s) on the command line.
"""

import os, sys

from Ft.Server.Server import FtServerServerException, Error
from Ft.Server.Core import Constants
from Ft.Server.Core.Lib import ConfigFile

def Run(serverNames=None, configFile=None, coreId=None):
    from Ft.Server.Protocols import Launcher
    properties = ConfigFile.ReadProperties(fileName=configFile)
    coreId = coreId or Constants.DEFAULT_CORE_ID

    serverDefs = {}
    for defs in properties.values():
        serverDefs.update(defs)

    status = {}
    if not serverNames:
        for name, prop in serverDefs.items():
            if prop.has_key('StartMode'):
                # It is a server entry
                try:
                    launcher = Launcher.Launcher(name, prop, display=1)
                    status[name] = launcher.status()
                except FtServerServerException, exc:
                    if exc.code != Error.NOT_SUPPORTED:
                        raise
    else:
        for name in serverNames:
            prop = serverDefs.get(name)
            if not prop:
                raise FtServerServerException(Error.UNKNOWN_SERVER, server=name)
            launcher = Launcher.Launcher(name, prop, display=1)
            status[name] = launcher.status()
    return status


def TextStatus(options, args):
    from Ft.Server.Protocols import Launcher
    try:
        serverNames = args['serverName']
        configFile = options.get('config-file')
        coreId = options.get('core-id')
        info = Run(serverNames, configFile, coreId)
    except FtServerServerException, e:
        print e
        return

    for serverName, (pid, status) in info.items():
        if status == Launcher.PID_RUNNING:
            print "%s (pid %d) is running..." % (serverName, pid)
        elif status in [Launcher.PID_DEAD, Launcher.PID_NOT_OWNER]:
            print "%s is dead but locked (pid %d)" % (serverName, pid)
        elif status == Launcher.PID_STOPPED:
            print "%s is stopped" % (serverName)
    return

def Register():
    from Ft.Lib.CommandLine import Options, Command, Arguments

    cmd = Command.Command('status',
                          "Display status of server(s)",
                          'PythonHttpServer',
                          __doc__,
                          function=TextStatus,
                          arguments = [Arguments.ZeroOrMoreArgument('serverName',
                                                                    'Name of server to wuery',
                                                                    str),
                                       ],
                          )
    return cmd



