"""
This HTTP handler is used to intercept a HTTP request uris to use to process and
return a search.  The trailing CGI funcion and variable (both specified in the configuration)
are then used in a request uri and the handler uses the variable for the search criteria.  For instance,
the following entries in a configuration file :

  <rdf:Description ID='SearchHandler'>
    <rdf:type resource='http://xmlns.4suite.org/4ss/properties#HttpHandler'/>
    <Module>Ft.Server.Protocols.Http.SearchHandler</Module>
    <AuthenticationRealm>4SSBrowser</AuthenticationRealm> 
    <QueryFunction>find</QueryFunction>  <!-- defaults to search ->
    <QueryVariable>words</QueryVariable> <!-- defaults to query ->
  </rdf:Description>

  ...

  <rdf:Description ID='SomeServer'>
    <rdf:type resource='http://xmlns.4suite.org/4ss/properties#PythonServer'/>
    <StartMode>AUTOMATIC</StartMode>
    <PidFile>/tmp/PyHttp.pid</PidFile>
    <Port>8080</Port>
    ... other handlers ...
    <Handler resource='#SearchHandler'/>
    <ErrorLog>/tmp/PyHttp.errors</ErrorLog>
    <TransferLog>/tmp/PyHttp.errors</TransferLog>
  </rdf:Description>
  

would have 4SS intercept the following request uri:

http://localhost:8080/find?words=Someword+and+anotherWord

would return a search results page w/ the results of running :
searchDocuments('Someword and anotherWord')

on any resource in the repository
"""




from CommonHandler import CommonHandler
from string import replace

RESULT_HTML="""
<HTML>
  <HEAD>
    <TITLE>Search results for :%s</TITLE>
  </HEAD>
  <BODY>
     Search Results for : %s<br>
     <OL>
     %s
     </OL>
  </BODY>
</HTML>
"""

class SearchHandler(CommonHandler):
    def __init__(self,  baseUri = "", serverName=None,rdfModelUri='',authenticationRealm=None,queryVar=''):
        self._serverName = serverName
        self._baseUri = baseUri
        self._rdfModelUri = rdfModelUri
        self.authenticationRealm = authenticationRealm  #If none we allow anonymous
        self.queryVar  = queryVar

    def handleSearch(self,request):
        kv_Dict=self._parseGETArguments(request)
        if kv_Dict.has_key(self.queryVar):
            query = kv_Dict[self.queryVar][0]
        else:
            return None
        query=replace(query,'+',' ')
        error,repo = self._do_authentication(request,self.authenticationRealm)
        if error:
            return error

        resultListItems=''
        resultPaths=repo.searchDocuments(query)
        print "query: %s, results: %s"%(query, resultPaths)
        for resultPath in repo.searchDocuments(query):
            resultListItems=resultListItems+"<LI>%s</LI\n"%(resultPath)
        resultBody=RESULT_HTML%(query,query,resultListItems)
        return self._sendNormalResponse(content=resultBody)
        

    pathMapping = {'search' : 'handleSearch'}    

def Register(props):
    authenticationRealm = props.get('AuthenticationRealm')
    queryFunc = props.get('QueryFunction','search')
    queryVar  = props.get('QueryVariable','query')
    

    g = SearchHandler(
        authenticationRealm=authenticationRealm,
        queryVar = queryVar)

    g.pathMapping[queryFunc]='handleSearch'
    return [(g.do_GET,'GET')]
