########################################################################
#
# File Name:            SoapApi.py
#
# Documentation:        http://docs.4suite.org/4SS/SoapApi.py.html
#
"""
Handles SOAP XML requests
WWW: http://4suite.org         e-mail: support@4suite.org

Copyright (c) 2000-2001 Fourthought, Inc. USA.   All Rights Reserved.
See  http://4suite.org/COPYRIGHT  for license and copyright information
"""

import string, base64, BaseHTTPServer, sys, cStringIO, traceback
from Ft.Xml import XPath
from Ft.Xml import Domlette
import CommonHandler
from Ft.Server.Common import ResourceTypes

FTSS_SOAP_NS = 'http://xmlns.4suite.org/services'

XMLSCHEMA_INSTANCE_NS = 'http://www.w3.org/1999/XMLSchema-instance'
SOAP_NS = "http://schemas.xmlsoap.org/soap/envelope/"
SOAP_ENCODING_STYLE = "http://schemas.xmlsoap.org/soap/encoding/"


def GenerateFault(respbody):
    eType, ex = sys.exc_info()[:2]

    doc = respbody.ownerDocument

    if 0:
        #FIXME Capture specific exceptions
        pass
    else:
        #Unknown Fault
        faultCode = "Server"
        faultString = str(ex)
        faultFactor = None
        st = cStringIO.StringIO()
        traceback.print_exc(file=st)
        detailStr = base64.encodestring(st.getvalue())
        detail = doc.createElementNS('','detail')
        tb = doc.createElementNS(FTSS_SOAP_NS,'ftsoap:traceback')
        tb.appendChild(doc.createTextNode(detailStr))
        detail.appendChild(tb)

    fault = doc.createElementNS(SOAP_NS,'SOAP-ENV:Fault')
    fc = doc.createElementNS('','faultCode')
    fc.appendChild(doc.createTextNode(str(faultCode)))
    fault.appendChild(fc)

    fs = doc.createElementNS('','faultString')
    fs.appendChild(doc.createTextNode(str(faultString)))
    fault.appendChild(fs)

    if faultFactor:
        ff = doc.createElementNS('','faultFactor')
        ff.appendChild(doc.createTextNode(str(faultFactor)))
        fault.appendChild(ff)

    if detail:
        fault.appendChild(detail)

    respbody.appendChild(fault)


def _MakeResourceResponse(res, respbody):
    doc = respbody.ownerDocument
    if res.isResourceType(ResourceTypes.ResourceType.CONTAINER):
        #Make a container response
        result = doc.createElementNS(FTSS_SOAP_NS,'ftsoap:Container')
        result.setAttributeNS('','uri',res.getUri())
    elif res.isResourceType(ResourceTypes.ResourceType.FILE):
        cont = res.getContent()
        imt = res.getImt()
        result = doc.createElementNS(FTSS_SOAP_NS,'ftsoap:Document')
        result.setAttributeNS('', 'uri', res.getUri())
        result.setAttributeNS('', 'imt', res.getImt())
        result.appendChild(doc.createTextNode(base64.encodestring(cont)))
    else:
        raise Exception("Finish")

    respbody.appendChild(result)

def FetchResource(repo, reqel, respbody):
    doc = respbody.ownerDocument
    context = XPath.Context.Context(reqel)
    #context = XPath.Context.Context(reqel, processorNss={'ft': XMLSERVER_NS})
    uri = XPath.Evaluate('string(@uri)', context=context)
    try:
        res = repo.fetchResource(uri, traverseAliases=1)
    except:
        GenerateFault(respbody)
        return 0
    _MakeResourceResponse(res,respbody)
    return 0


def CreateResource(repo, reqel, respbody):
    doc = respbody.ownerDocument
    context = XPath.Context.Context(reqel, processorNss={'ft': FTSS_SOAP_NS})
    cont = XPath.Evaluate('ft:Container', context=context)
    ddName = XPath.Evaluate('string(@doc-def)', context=context)

    if cont:
        context = XPath.Context.Context(cont[0], processorNss={'ft': FTSS_SOAP_NS})
        uri = XPath.Evaluate('string(@uri)', context=context)
        cp = XPath.Evaluate('number(@create-parents)', context=context)
        try:
            res = repo.createContainer(uri, createParents=cp)
        except:
            GenerateFault(respbody)
            return 0
    elif ddName:
        uri = XPath.Evaluate('string(@uri)', context=context)
        source = base64.decodestring(XPath.Evaluate('string(node())',
                                                    context=context))
        res = repo.createDocument(ddName, uri, source)
    else:
        try:
            raise "Unknown Resource Type"
        except:
            GenerateFault(respbody)
            return 0
    _MakeResourceResponse(res,respbody)
    return 1


def DeleteResource(repo, reqel, respbody):
    doc = respbody.ownerDocument
    context = XPath.Context.Context(reqel)
    #context = XPath.Context.Context(reqel, processorNss={'ft': XMLSERVER_NS})
    uri = XPath.Evaluate('string(@uri)', context=context)
    try:
        repo.deleteResource(uri)
    except:
        GenerateFault(respbody)
        return 0

    return 1


class SoapHandler(CommonHandler.CommonHandler):
    reader = Domlette.DEFAULT_NONVALIDATING_READER()

    NS_TO_HANDLER_MAPPING = {
        FTSS_SOAP_NS: {'Fetch': FetchResource,
                       'Create': CreateResource,
                       'Delete': DeleteResource,
                       },
        }

    def handle(self, request):
        return self._safeHandleRequest(self._handle, request)

    def _handle(self, request):
        for name in request.headers.keys():
            if string.lower(name) == 'soapaction':
                break
        else:
            return None
        doc = self.reader.fromString(request.body)

        handlerfunc, retval = self._checkHeader(request, doc)
        if not handlerfunc:
            return retval
        reqbody = retval
        (error, repo) = self._do_authentication(request,
                                                self.authenticationRealm)
        if error:
            # error is a response object
            return error

        try:
            commit = 0
            respbody = self._prepResponse(request, reqbody)
            #Envoke our handler
            commit = handlerfunc(repo, reqbody, respbody)
            return self._finalizeResponse(request, reqbody, respbody)
        finally:
            if commit:
                repo.txCommit()
            else:
                repo.txRollback()

    def _checkHeader(self, request, doc):
        #Read in the SOAP request envelope
        soap_envelope = doc.documentElement
        if (soap_envelope.namespaceURI, soap_envelope.localName) != (SOAP_NS, "Envelope"):
            #FIXME: Use HTTP SoapAction header for screening insead.  If it's present but the above condition is false, we send a client fault
            return None, None
            #FIXME: This should return a SOAP client fault
            #raise Exception('Invalid SOAP message')
        soap_body = filter(lambda x: (x.namespaceURI, x.localName) == (SOAP_NS, "Body"), soap_envelope.childNodes)[0]
        request_el = filter(lambda x: x.nodeType == Node.ELEMENT_NODE, soap_body.childNodes)[0]
        request_ns = request_el.namespaceURI
        request_name = request_el.localName

        #Determine our handler
        try:
            handlerfunc1 = self.NS_TO_HANDLER_MAPPING[request_ns]
        except KeyError:
            raise Exception("No handler for SOAP namespace: " + request_ns)

        try:
            handlerfunc = handlerfunc1[request_name]
        except KeyError:
            raise Exception("No handler for SOAP request: " + request_name)

        if not request.body:
            raise Exception('POST ERROR: missing body')

        return handlerfunc, request_el

    def _prepResponse(self, request, doc):
        #Create the response document
        dt = implementation.createDocumentType("SOAP-ENV:Envelope", '', '')
        respdoc = implementation.createDocument(SOAP_NS, "SOAP-ENV:Envelope", dt)
        respdocel = respdoc.documentElement
        respdocel.setAttributeNS(SOAP_NS, "SOAP-ENV:encodingStyle", SOAP_ENCODING_STYLE)
        respbody = respdoc.createElementNS(SOAP_NS, "SOAP-ENV:Body")
        respdocel.appendChild(respbody)
        return respbody

    def _finalizeResponse(self, request, reqbody, respbody):
        #Serialize the results
        stream = cStringIO.StringIO()
        ext.PrettyPrint(respbody.ownerDocument, stream=stream)
        bodystr = stream.getvalue()

        #Send our response
        return CommonHandler.Response(
            200, {'Content-Type':'text/html; charset=utf-8'}, bodystr
            )


def Register(props):
    #FIXME: Implement authentication through SOAP headers
    handler = SoapHandler()
    return [(handler.handle, 'POST')]

