#include "domlette.h"

#define NODEVALUE_CHECK(comment) if (!PyComment_Check((comment)) || !((PyCommentObject *)(comment))->nodeValue) {PyErr_SetString(g_errorObject, "Comment in inconsistent state"); return NULL;}
#define COMMENT_CHECK(comment) NODEVALUE_CHECK((comment));

PyObject *g_commentNodeType; // added by Brendan Larson 9/12/2002 for Mac OS X 10.1.x compiler compatiblity
PyObject *g_commentNodeName; //   IBID

/*Internal Methods*/
PyCommentObject *Comment_CloneNode(PyCommentObject *node,int deep,PyObject *newOwnerDocument) {

  long tempId;
  PyObject *nodeValue;

  if (!PyDocument_Check(newOwnerDocument)) {
    PyErr_SetString(PyExc_TypeError,"Argument must be a document");
    return NULL;
  }
  nodeValue = PyObject_GetAttrString((PyObject *)node,"nodeValue");
  if (!nodeValue) return NULL;
  return Document_CreateComment((PyDocumentObject *)newOwnerDocument, nodeValue, &tempId);
}


/*External Methods*/


static struct PyMethodDef Comment_methods[] = { 
  NODE_METHODS,
  CHARACTERDATA_METHODS,
  {NULL, NULL}      /* sentinel */
};



/*Type Methods*/

static PyObject *comment_getattr(PyCommentObject *self, char *name)
{
  PyObject *rt = NULL;

  COMMENT_CHECK(self);

  if (!strcmp(name,"data")) {
    rt = (PyObject *)self->nodeValue;
  }
  else if (!strcmp(name,"nodeValue")) {
    rt = (PyObject *)self->nodeValue;
  }
  else if (!strcmp(name,"nodeType")) {
    rt = g_commentNodeType;
  }
  else if (!strcmp(name,"nodeName")) {
    rt = g_commentNodeName;
  }
  if (rt) {
    Py_INCREF(rt);
    return rt;
  }

  return node_getattr((PyNodeObject*)self,name,Comment_methods);
}


static void comment_dealloc(PyCommentObject *node)
{
  PyObject_GC_UnTrack((PyObject *) node);

  Py_XDECREF(node->nodeValue);
  node->nodeValue = NULL;

  node_dealloc((PyNodeObject*)node);
}

static int comment_clear(PyCommentObject *self)
{
  /*Called when we need to break cycles*/
  return node_clear((PyNodeObject *)self);
}

static int comment_traverse(PyCommentObject *self, visitproc visit, void *arg)
{
  return node_traverse((PyNodeObject *)self,visit,arg);
}



static PyObject *comment_repr(PyCommentObject *comment)
{
  char buf[256];
  sprintf(buf, "<cComment at %p>", comment);
  return PyString_FromString(buf);
}

static int comment_print(PyCommentObject *v, FILE *fp, int flags)
/* flags not used but required by interface */
{
  PyObject *repr = comment_repr(v);
  fprintf(fp, PyString_AS_STRING(repr));
  Py_DECREF(repr);
  return 0;
}

int comment_test_ref_counts(PyObject *tester,PyCommentObject *node,long *childCtr,PyObject *internDict,int base) {
  /*Ref count is base + 1.  Out nodeValue should have a ref count of 1 but is not interned*/
  PyObject_CallMethod(tester,"startTest","s","Node");
  if (!PyObject_CallMethod(tester,"compare","ll",base+1,node->ob_refcnt)) return 0;
  PyObject_CallMethod(tester,"testDone","");
  if (!TestRefCount(tester,(PyObject *)node->nodeValue,1, "nodeValue")) return 0;
  return 1;
}

PyTypeObject PyDomletteComment_Type = {
    PyObject_HEAD_INIT(0)
    0,
    "cDomlette.Comment",
    GC_TP_BASICSIZE(PyCommentObject),
    0,
    (destructor)comment_dealloc,    /*tp_dealloc*/
    (printfunc)comment_print,   /*tp_print*/
    (getattrfunc)comment_getattr,   /*tp_getattr*/
    (setattrfunc)characterdata_setattr,  /*tp_setattr*/
    0,                          /*tp_compare*/
    (reprfunc)comment_repr,          /*tp_repr*/
    0,                          /*tp_as_number*/
    0,              /*tp_as_sequence*/
    0,              /*tp_as_mapping*/
    0,                             /*tp_hash*/
    0,          /*tp_call*/
    0,          /*tp_str*/
    0,                      /*tp_getattro*/
    0,          /*tp_setattro*/
    0,                          /* tp_as_buffer */
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,                          /* tp_doc */
    (traverseproc)comment_traverse,  /* tp_traverse */
    (inquiry)comment_clear,          /* tp_clear */
    0,                          /* tp_richcompare */
    0,                          /* tp_weaklistoffset */
};

