#include "domlette.h"

PyObject *g_cdataSectionNodeType; // added by Brendan Larson 9/12/2002 for Mac OS X 10.1.x compiler compatiblity
PyObject *g_entityReferenceNodeType; // added by Brendan Larson 9/12/2002 for Mac OS X 10.1.x compiler compatiblity
PyObject *g_entityNodeType; // added by Brendan Larson 9/12/2002 for Mac OS X 10.1.x compiler compatiblity
PyObject *g_notationNodeType; // added by Brendan Larson 9/12/2002 for Mac OS X 10.1.x compiler compatiblity
PyObject *g_cdataSectionNodeName; // added by Brendan Larson 9/12/2002 for Mac OS X 10.1.x compiler compatiblity

long g_nodeCount = 0;


#define PARENTNODE_CHECK(node) if (!PyNode_Check((node)) || !((PyNodeObject *)(node))->parentNode) {PyErr_SetString(g_errorObject, "Node in inconsistent state"); return NULL;}
#define OWNERDOCUMENT_CHECK(node) if (!PyNode_Check((node)) || !((PyNodeObject *)(node))->ownerDocument) {PyErr_SetString(g_errorObject, "Node in inconsistent state"); return NULL;}
#define NODE_CHECK(node) PARENTNODE_CHECK((node));\
OWNERDOCUMENT_CHECK((node));


/*Initialize the a new Node*/
void Node_INIT(PyNodeObject *node, PyObject *ownerDocument,
               PyTypeObject *type
               )
{
  node->parentNode = Py_None;
  Py_XINCREF(Py_None);

  node->ownerDocument = (PyObject*)ownerDocument;
  Py_INCREF(ownerDocument);

  node->ob_type = type;

  g_nodeCount++;
#ifdef DEBUG_NODE_CREATION
  printf("Created Node (%s): %p\n", node->ob_type->tp_name, node);
#endif
}


PyNodeObject *Node_RemoveChild(PyNodeObject *self, PyNodeObject *removeChild) {

  /*
  Remove the child from its parent list
  Assuming the child is actually a child of the parent, find out its index
  then, remove it from our list of children.
  DECREF the child once to remove the dependency of the parent to the child
  DECREF the parent once to remove the dependency of the child to the parent
  set the child's parent Node to Py_None and INCREF Py_None
  */

  PyObject *children;
  int index, ctr;

  NODE_CHECK(self);

#ifdef DEBUG_NODE_REMOVE_CHILD
  printf("Remove Child initial state\n");
  printf("self\n");
  _PyObject_Dump(self);
  printf("------------------\n");
  printf("remove child\n");
  _PyObject_Dump(removeChild);
  printf("------------------\n");
#endif

  /*Only docs, elements and dfs*/
  if (PyDocument_Check(self)) {
    children = ((PyDocumentObject *)self)->childNodes; 
  } else if (PyElement_Check(self)){
    children = ((PyElementObject *)self)->childNodes; 
  } else if (PyDocumentFragment_Check(self)){
    children = ((PyDocumentFragmentObject *)self)->childNodes; 
  } else {
    PyErr_SetString(g_hierarchyRequestErr,"Not allowed to have children");
    return NULL;
  }

#ifdef DEBUG_NODE_REMOVE_CHILD
  printf("children\n");
  _PyObject_Dump(children);
  printf("------------------\n");
#endif
  
  /*Find the index of the child to be removed*/
  index = -1;
  for (ctr=0;ctr <  PyList_GET_SIZE(children);ctr++) {
    if (PyList_GET_ITEM(children, ctr) == (PyObject *)removeChild) {
      index = ctr;
      break;
    }
  }

  if (index == -1) {
    PyErr_SetString(g_notFoundErr,"Child not found");
    return NULL;
  }
  
#ifdef DEBUG_NODE_REMOVE_CHILD
  printf("index: %d\n",index);
#endif

  /*We have an index, call PySequence_DelItem which will DECREF child*/
  /*Temporarily INCREF the remove child incase the list has the last reference to it*/
  Py_INCREF(removeChild);

  PySequence_DelItem(children, index);

#ifdef DEBUG_NODE_REMOVE_CHILD
  printf("removeChild after delItem\n");
  _PyObject_Dump(removeChild);
  printf("------------------\n");
#endif
 
  /*Reset the parentNode on the child to Py_None, DECREF parent Node and INCREF Py_None*/
  Py_DECREF(removeChild->parentNode);
  removeChild->parentNode = Py_None;
  Py_INCREF(Py_None);

#ifdef DEBUG_NODE_REMOVE_CHILD
  printf("self after parent change\n");
  _PyObject_Dump(self);
  printf("------------------\n");
#endif

  if (PyDocument_Check(self) && PyElement_Check(removeChild)) {
    /*If we made it this far then it had to be the doc elem*/
    Py_DECREF(((PyDocumentObject *)self)->documentElement);
    ((PyDocumentObject *)self)->documentElement = Py_None;
    Py_XINCREF(Py_None);
  }      

  /*Undo the temp INCREF*/
  Py_DECREF(removeChild);

#ifdef DEBUG_NODE_REMOVE_CHILD
  printf("Remove Child final state\n");
  printf("self\n");
  _PyObject_Dump(self);
  printf("------------------\n");
  printf("remove child\n");
  _PyObject_Dump(removeChild);
  printf("------------------\n");
  printf("children\n");
  _PyObject_Dump(children);
  printf("------------------\n");
#endif

  return removeChild;
}


PyNodeObject *Node_AppendChild(PyNodeObject *self, PyNodeObject *child)
{
  PyObject *children;
  PyObject *dfChildren;

  NODE_CHECK(self);
  NODE_CHECK(child);

#ifdef DEBUG_NODE_APPEND_CHILD
  printf("Append Child initial state\n");
  printf("self\n");
  _PyObject_Dump(self);
  printf("------------------\n");
  printf("child\n");
  _PyObject_Dump(child);
  printf("------------------\n");
#endif

  if PyDocument_Check(self)
  {
    children = ((PyDocumentObject *)self)->childNodes;
  } else if PyElement_Check(self) {
    children = ((PyElementObject *)self)->childNodes;
  } else if PyDocumentFragment_Check(self) {
    children = ((PyDocumentFragmentObject *)self)->childNodes;
  } else {
    PyErr_SetString(g_hierarchyRequestErr,"Invalid Node Type");
    return NULL;
  }

#ifdef DEBUG_NODE_APPEND_CHILD
  printf("children\n");
  _PyObject_Dump(children);
  printf("------------------\n");
#endif

  if (PyDocument_Check(self) && PyElement_Check(child)){
    if (PyElement_Check(((PyDocumentObject *)self)->documentElement)) {
      /*raise hiearchy error*/
      PyErr_SetString(g_hierarchyRequestErr,"two document elements disallowed");
      return NULL;
    }
#ifdef DEBUG_NODE_APPEND_CHILD
    printf("Appending a new document element\n");
#endif
    Py_DECREF(((PyDocumentObject *)self)->documentElement);
    ((PyDocumentObject *)self)->documentElement = (PyObject *)child;
    Py_INCREF(child);
  } 
  if (PyDocumentFragment_Check(child)) {
    int i, size;
#ifdef DEBUG_NODE_APPEND_CHILD
    printf("Append a Document Fragment\n");
#endif
    /*Loop and add them all Removeing them from the df*/
    dfChildren = ((PyDocumentFragmentObject *)child)->childNodes;
    size = PyList_GET_SIZE(dfChildren);
    for (i = 0; i < size; i++) {
      /*Call append child on each of the children, append child does a remove if needed*/
      PyObject *dfChild = PyList_GET_ITEM(dfChildren, 0);
#ifdef DEBUG_NODE_APPEND_CHILD
      printf("dfChild %d before sub append\n",i);
      _PyObject_Dump(dfChild);
      printf("------------------\n");
#endif
      Node_AppendChild(self,(PyNodeObject *)dfChild);
#ifdef DEBUG_NODE_APPEND_CHILD
      printf("dfChild %d after sub append\n",i);
      _PyObject_Dump(dfChild);
      printf("------------------\n");
#endif
    }
  } else {
    
    /*Do a simple append
      If the node already has a parent node, we need to remove it from its parent.
      Set its parent node to us
      Add it to the end of our child list

      Always incref the child once to create our own reference to it
      Always incref ourselves once because of the parent node relationship

    */


    /*Do a temp INCREF incase the node loses its only reference in the remove child*/
    Py_INCREF(child);

    if (child->parentNode == Py_None) {
      /*No current parent, decref Py_None*/
      Py_DECREF(Py_None);
    } else {
      /*Child has a parent, remove the child from the parent, remove Child does a decref of parent*/
      Node_RemoveChild((PyNodeObject *)child->parentNode,(PyNodeObject *)child);
    }

    /*Now, add the node to our list of children.  This INCs the node once*/
    PyList_Append(children,(PyObject *)child);

#ifdef DEBUG_NODE_APPEND_CHILD
    printf("child after list append\n");
    _PyObject_Dump(child);
    printf("------------------\n");
#endif
    /*Set the parent node to us, old parent already DECREFed*/
    child->parentNode = (PyObject *)self;
    Py_INCREF(self);

    /*Undo the temp INCREF*/
    /*Logic tells me that I should DECREF this, but if I do things get very wierd*/
    Py_DECREF(child);

  }

#ifdef DEBUG_NODE_APPEND_CHILD
  printf("Append Child final state\n");
  printf("self\n");
  _PyObject_Dump(self);
  printf("------------------\n");
  printf("child\n");
  _PyObject_Dump(child);
  printf("------------------\n");
  printf("children\n");
  _PyObject_Dump(children);
  printf("------------------\n");
#endif

  return child;
}



PyNodeObject *Node_insertBefore(PyNodeObject *self, PyNodeObject *newChild, PyNodeObject *refChild)
{
  PyObject *children;
  PyObject *dfChildren, *res;
  PyObject *dfChild;
  int ctr, index;

  NODE_CHECK(self);

#ifdef DEBUG_NODE_INSERT_BEFORE
  printf("Starting Insert Before\n");
  printf("new child\n");
  _PyObject_Dump(newChild);
  printf("---------------\n");
  printf("ref child\n");
  _PyObject_Dump(refChild);
  printf("---------------\n");
  printf("self\n");
  _PyObject_Dump(self);
  printf("---------------\n");
#endif

  if PyDocument_Check(self)
  {
    children = ((PyDocumentObject *)self)->childNodes;
  } else if PyElement_Check(self) {
    children = ((PyElementObject *)self)->childNodes;
  } else if PyDocumentFragment_Check(self) {
    children = ((PyDocumentFragmentObject *)self)->childNodes;
  } else {
    PyErr_SetString(g_notFoundErr,"Reference node not found");
    return NULL;
  }

#ifdef DEBUG_NODE_INSERT_BEFORE
  printf("children\n");
  _PyObject_Dump(children);
  printf("---------------\n");
#endif

  if ((PyObject *)refChild == Py_None) {
#ifdef DEBUG_NODE_INSERT_BEFORE
    printf("Doing an append\n");
#endif
    res = (PyObject *)Node_AppendChild((PyNodeObject *)self,(PyNodeObject *)newChild);
    if (res == NULL) return NULL;
  } else {
    if (PyDocument_Check(self) && PyElement_Check(newChild))
      {
	if (PyElement_Check(((PyDocumentObject *)self)->documentElement)) {
	  /*raise hiearchy error*/
	  PyErr_SetString(g_hierarchyRequestErr,"Only one document element allowed");
	  return NULL;
	}
#ifdef DEBUG_NODE_INSERT_BEFORE
	printf("Setting Owner Document\n");
#endif
	Py_DECREF(((PyDocumentObject *)self)->documentElement);
	((PyDocumentObject *)self)->documentElement = (PyObject *)newChild;
	Py_INCREF(newChild);
      } 

    /*Find the index*/
    index = -1;
    for (ctr=0;ctr <  PyList_GET_SIZE(children);ctr++) {
      if (PyList_GET_ITEM(children,ctr) == (PyObject *)refChild) {
        index = ctr;
        break;
      }
    }
    if (index == -1) {
      PyErr_SetString(g_notFoundErr,"Reference Node not found");
      return NULL;
    }
#ifdef DEBUG_NODE_INSERT_BEFORE
    printf("Index of %d\n",index);
#endif
    
    if (PyDocumentFragment_Check(newChild)) {
      int i, size;
      /*Loop and add them all Removeing them from the df*/
      dfChildren = ((PyDocumentFragmentObject *)newChild)->childNodes;
      size = PyList_GET_SIZE(dfChildren);
      for (i = 0; i < size; i++) {
#ifdef DEBUG_NODE_INSERT_BEFORE
	printf("*****Start insert of DF child %d*****\n",i);
#endif
	dfChild = PyList_GET_ITEM(dfChildren, 0);
	Node_insertBefore(self,(PyNodeObject *)dfChild,refChild);
#ifdef DEBUG_NODE_INSERT_BEFORE
	printf("*****End insert of DF child %d*****\n",i);
#endif
      }
      
    } else {


      /*If the node has a parent node that is not Py_None, we need to remove them from
	that parent.
	Then, insert them into our list, doing a single INCREF to the new child.
	
      */

      /*Do a temp INCREF incase the node loses its only reference in the remove child*/
      Py_INCREF(newChild);

      if (newChild->parentNode == Py_None) {
	/*No current parent, decref Py_None*/
	Py_DECREF(Py_None);
      } else {
	/*Child has a parent, remove the child from the parent, remove Child does a decref of parent*/
	Node_RemoveChild((PyNodeObject *)newChild->parentNode,(PyNodeObject *)newChild);
      }

      /*Now, insert the node into our list at the correct location*/
      PyList_Insert(children,index,(PyObject *)newChild);

#ifdef DEBUG_NODE_INSERT_BEFORE
      printf("newChild after list append\n");
      _PyObject_Dump(newChild);
      printf("------------------\n");
#endif
      /*Set the parent node to us, old parent already DECREFed*/
      newChild->parentNode = (PyObject *)self;
      Py_INCREF(self);

      /*Undo the temp INCREF*/
      Py_DECREF(newChild);
    }
  }


#ifdef DEBUG_NODE_INSERT_BEFORE
  printf("Ending Insert Before\n");
  printf("new child\n");
  _PyObject_Dump(newChild);
  printf("---------------\n");
  printf("ref child\n");
  _PyObject_Dump(refChild);
  printf("---------------\n");
  printf("self\n");
  _PyObject_Dump(self);
  printf("---------------\n");
#endif
  return newChild;
}


PyNodeObject *Node_CloneNode(PyNodeObject *node, int deep, PyObject *newOwnerDocument) {

  PyObject *obj;
  int node_type;

  /* Get the nodeType as a plain integer */
  obj = PyObject_GetAttrString((PyObject *)node, "nodeType");
  if (!obj) return NULL;
  node_type = PyInt_AsLong(obj);
  Py_DECREF(obj);

  switch (node_type) {
  case ELEMENT_NODE:
    return (PyNodeObject *)Element_CloneNode((PyElementObject *)node, deep, 
					     newOwnerDocument);
  case TEXT_NODE:
    return (PyNodeObject *)Text_CloneNode((PyTextObject *)node, deep, 
					  newOwnerDocument);
  case COMMENT_NODE:
    return (PyNodeObject *)Comment_CloneNode((PyCommentObject *)node, deep, 
					     newOwnerDocument);
  case DOCUMENT_FRAGMENT_NODE:
    return (PyNodeObject *)DocumentFragment_CloneNode((PyDocumentFragmentObject *)node,
						      deep, newOwnerDocument);
  case PROCESSING_INSTRUCTION_NODE:
    return (PyNodeObject *)ProcessingInstruction_CloneNode((PyProcessingInstructionObject *)node,
							   deep, newOwnerDocument);
  default:
    PyErr_Format(g_errorObject, "cloneNode: unknown nodeType %d", node_type);
    return NULL;
  }
}


/*
    Start the external interface

*/



PyObject *PyNode_normalize(PyObject * self, PyObject * args) {


  PyObject *children;
  int ctr;
  PyObject *current, *next, *temp;

  NODE_CHECK(self);

#ifdef DEBUG_NODE_NORMALIZE
  printf("normalize initial state\n");
  printf("self\n");
  _PyObject_Dump(self);
  printf("------------------\n");
#endif

  if PyDocument_Check(self)
  {
    children = ((PyDocumentObject *)self)->childNodes;
  } else if PyElement_Check(self) {
    children = ((PyElementObject *)self)->childNodes;
  } else if PyDocumentFragment_Check(self) {
    children = ((PyDocumentFragmentObject *)self)->childNodes;
  } else {
    Py_INCREF(Py_None);
    return Py_None;
  }
  if (PyList_GET_SIZE(children) < 2) {
    Py_INCREF(Py_None);
    return Py_None;
  }

#ifdef DEBUG_NODE_NORMALIZE
  printf("children\n");
  _PyObject_Dump(children);
  printf("------------------\n");
#endif

  ctr = 0;
  /*Count from 0 to 1 minus the length (the last node cannot be normalized with anything*/
  while (ctr < PyList_GET_SIZE(children)-1) {
    /*See if the current node is a text node*/
    current = PyList_GET_ITEM(children,ctr);
#ifdef DEBUG_NODE_NORMALIZE
    printf("Current Node\n");
    _PyObject_Dump(current);
    printf("------------------\n");
#endif
    if (PyText_Check(current)) {
      next = PyList_GET_ITEM(children,ctr+1);
#ifdef DEBUG_NODE_NORMALIZE
      printf("Next Node\n");
      _PyObject_Dump(next);
      printf("------------------\n");
#endif
      if (PyText_Check(next)) {
	/*Merge them*/
	temp = PySequence_Concat((PyObject *)((PyTextObject *)current)->nodeValue, (PyObject *)((PyTextObject *)next)->nodeValue);
#ifdef DEBUG_NODE_NORMALIZE
	printf("Merged nodeValue\n");
	_PyObject_Dump(temp);
	printf("------------------\n");
#endif
	
	Py_DECREF(((PyTextObject *)current)->nodeValue);
	((PyTextObject *)current)->nodeValue = temp;
	
	/*Remove it*/
	Node_RemoveChild((PyNodeObject *)self,(PyNodeObject *)next);
#ifdef DEBUG_NODE_NORMALIZE
	printf("self after merge\n");
	_PyObject_Dump(self);
	printf("------------------\n");
	printf("current after merge\n");
	_PyObject_Dump(current);
	printf("------------------\n");
#endif
      } else {
	ctr ++;
      }
    } else {
      ctr++;
    }
  }

#ifdef DEBUG_NODE_NORMALIZE
  printf("normalize final state\n");
  printf("self\n");
  _PyObject_Dump(self);
  printf("------------------\n");
  printf("children\n");
  _PyObject_Dump(children);
  printf("------------------\n");
#endif

  Py_INCREF(Py_None);
  return Py_None;
}

PyObject *PyNode_hasChildNodes(PyObject * self, PyObject * args) {

  PyObject *rt;

  NODE_CHECK(self);

  if (PyDocument_Check(self) && PyList_GET_SIZE(((PyDocumentObject *)self)->childNodes)) {
    rt = Py_True;
  } else if(PyElement_Check(self) && PyList_GET_SIZE(((PyElementObject *)self)->childNodes)) {
    rt = Py_True;
  } else if(PyAttr_Check(self)) {
    rt = Py_True;
  } else {
    rt = Py_False;
  }
  Py_INCREF(rt);
  return rt;
}

PyObject *PyNode_removeChild(PyObject *self,PyObject *args) {

  PyObject *removeChild, *res;

  NODE_CHECK(self);

  if(!PyArg_ParseTuple(args,"O:removeChild",&removeChild)) 
    return NULL;

  res = (PyObject *)Node_RemoveChild((PyNodeObject *)self,(PyNodeObject *)removeChild);

  if (res == NULL) return res;
  /*INCREF the results for the return value reference*/
  Py_INCREF(res);
  return res;
}


PyObject *PyNode_isSameNode(PyObject *self, PyObject *args)
{
  PyObject *other;

  NODE_CHECK(self);

  if(!PyArg_ParseTuple(args,"O:other", &other)) 
    return NULL;

  return PyInt_FromLong(self == other);
}

PyObject *PyNode_appendChild(PyObject *self, PyObject *args)
{
  PyObject *newChild;
  PyNodeObject *oldChild;

  NODE_CHECK(self);

  if(!PyArg_ParseTuple(args,"O:appendChild", &newChild)) 
    return NULL;

  /*Append child adds a ref to the child to signify the new relationship*/
  oldChild = Node_AppendChild((PyNodeObject *)self,(PyNodeObject *)newChild);
  if (oldChild == NULL) return NULL;
  /*Add a ref to the child for the return value*/
  Py_INCREF(oldChild);
  return (PyObject *)oldChild;
}


PyObject *PyNode_insertBefore(PyObject *self, PyObject *args)
{
  PyObject *newChild, *refChild;

  NODE_CHECK(self);

  if(!PyArg_ParseTuple(args,"OO:insertBefore", &newChild,&refChild)) 
    return NULL;

  if (!Node_insertBefore((PyNodeObject *)self,(PyNodeObject *)newChild,(PyNodeObject *)refChild)) {
    return NULL;
  }
  /*INCREF newChild for the return value*/
  Py_INCREF(newChild);
  return newChild;
}


PyObject *PyNode_replaceChild(PyObject *self, PyObject *args)
{
  PyObject *newChild, *oldChild, *sibling;

  struct PyMethodDef Bogus_methods[] = {
  {NULL,     NULL}      /* sentinel */
  };

  NODE_CHECK(self);

  if(!PyArg_ParseTuple(args,"OO:replaceChild", &newChild,&oldChild)) 
    return NULL;

#ifdef DEBUG_NODE_REPLACE_CHILD
  printf("replace child initial state\n");
  printf("self\n");
  _PyObject_Dump(self);
  printf("------------------\n");
  printf("newChild\n");
  _PyObject_Dump(newChild);
  printf("------------------\n");
  printf("oldChild\n");
  _PyObject_Dump(oldChild);
  printf("------------------\n");
#endif

  sibling = node_getattr((PyNodeObject *)oldChild,"nextSibling",Bogus_methods);

#ifdef DEBUG_NODE_REPLACE_CHILD
  printf("sibling\n");
  _PyObject_Dump(sibling);
  printf("------------------\n");
#endif

  /*INC the return value before we call remove incase this is the last reference*/
  Py_INCREF(oldChild);

  if (!Node_RemoveChild((PyNodeObject *)self,(PyNodeObject *)oldChild)) {
    return NULL;
  }

#ifdef DEBUG_NODE_REPLACE_CHILD
  printf("oldChild after remove\n");
  _PyObject_Dump(oldChild);
  printf("------------------\n");
#endif

  /*Now insert it before the sibling (sibling could be Py_None but that is OK)*/
  if (!Node_insertBefore((PyNodeObject *)self,(PyNodeObject *)newChild,(PyNodeObject *)sibling)) {
    return NULL;
  }

#ifdef DEBUG_NODE_REPLACE_CHILD
  printf("newChild after insert\n");
  _PyObject_Dump(oldChild);
  printf("------------------\n");
#endif

  /*DECREF the sibling (getattr added one)*/
  Py_DECREF(sibling);

#ifdef DEBUG_NODE_REPLACE_CHILD
  printf("replace child final state\n");
  printf("self\n");
  _PyObject_Dump(self);
  printf("------------------\n");
  printf("newChild\n");
  _PyObject_Dump(newChild);
  printf("------------------\n");
  printf("oldChild\n");
  _PyObject_Dump(oldChild);
  printf("------------------\n");
  printf("sibling\n");
  _PyObject_Dump(oldChild);
  printf("------------------\n");
#endif
  return oldChild;
}


PyObject *PyNode_cloneNode(PyObject *self, PyObject *args)
{
  int deep = 0;

  NODE_CHECK(self);

  if(!PyArg_ParseTuple(args,"|i:cloneNode", &deep)) 
    return NULL;

  if (PyDocument_Check(self)) {
    PyErr_SetString(PyExc_TypeError,"cloneNode not allowed on documents");
    return NULL;
  } 
  return (PyObject *)Node_CloneNode((PyNodeObject *)self,deep,((PyNodeObject *)self)->ownerDocument);
}


/*Used by the all of the types as helpers */
void node_dealloc(PyNodeObject *node)
{
#ifdef DEBUG_NODE_CREATION
  printf("Destroyed Node (%s): %p\n",node->ob_type->tp_name,node);
#endif

  g_nodeCount--;
  Py_XDECREF(node->parentNode);
  node->parentNode = NULL;
  Py_XDECREF(node->ownerDocument);
  node->ownerDocument = NULL;
  PyObject_GC_Del((PyObject *)node);
}

int node_clear(PyNodeObject *node)
{
#ifdef DEBUG_NODE_CREATION
  printf("Clearing Node (%s): %p\n",node->ob_type->tp_name,node);
#endif
  Py_XDECREF(node->parentNode);
  node->parentNode = NULL;
  Py_XDECREF(node->ownerDocument);
  node->ownerDocument = NULL;
  return 0;
}

int node_traverse(PyNodeObject *self, visitproc visit, void *arg)
{
  int rt;
#ifdef DEBUG_NODE_CREATION
  printf("Traversing Node (%s): %p\n",self->ob_type->tp_name,self);
#endif
  if (self->parentNode != NULL) {
    rt = visit(self->parentNode, arg);
    if (rt) return rt;
  }
  if (self->ownerDocument != NULL) {
    rt = visit(self->ownerDocument, arg);
    if (rt) return rt;
  }
  return 0;
}


PyObject *node_getattr(PyNodeObject *self, char *name,struct PyMethodDef methods[])
{
  PyObject * rt = NULL;
  int temp;
  PyObject *pNode = NULL;
  int pNodeIndex = -1;
  PyObject *childNodes;

  NODE_CHECK(self);

  switch(name[0]) {
  case 'a':
    {
      if (!strcmp(name,"attributes")) {
	rt = g_emptyAttributes;
      }
      break;
    }
  case 'c':
    {
      if (!strcmp(name,"childNodes")) {
	return PyList_New(0);
      }
      break;
    }
  case 'f':
    {
      if (!strcmp(name,"firstChild")) {
        rt = Py_None;
      }
      break;
    }
  case 'l':
    {
      if (!strcmp(name,"localName")) {
        rt = Py_None;
      }
      else if (!strcmp(name,"lastChild")) {
        rt = Py_None;
      }
      break;
    }
  case 'n':
    {
      if (!strcmp(name,"nodeValue")) {
        rt = Py_None;
      }
      else if (!strcmp(name,"namespaceURI")) {
        rt = Py_None;
      }
      else if (!strcmp(name,"nextSibling")) {
	/*Calculate this*/
	pNode = self->parentNode;
	if (PyDocument_Check(pNode)) {
	  childNodes = ((PyDocumentObject *)pNode)->childNodes;
	} else if (PyElement_Check(pNode)) {
	  childNodes = ((PyElementObject *)pNode)->childNodes;
	} else if (PyDocumentFragment_Check(pNode)) {
	  childNodes = ((PyDocumentFragmentObject *)pNode)->childNodes;
	} else {
	  Py_INCREF(Py_None);
	  return Py_None;
	}
	for (temp = 0;temp < PyList_GET_SIZE(childNodes);temp ++) {
	  if ((PyNodeObject *)PyList_GET_ITEM(childNodes,temp) == self) {
	    pNodeIndex = temp;
	    break;
	  }
	}
	if (pNodeIndex == -1) {
	  /*Through error????*/
	  Py_INCREF(Py_None);
	  return Py_None;
	}
	if (pNodeIndex == PyList_GET_SIZE(childNodes) - 1) {
	  /*End of the list*/
	  rt = Py_None;
	} else {
	  rt = PyList_GET_ITEM(childNodes,pNodeIndex+1);
	}
      }
      break;
    }
  case 'o':
    {
      if (!strcmp(name,"ownerDocument")) {
        rt = self->ownerDocument;
      }
      break;
    }
  case 'p':
    {
      if (!strcmp(name,"prefix")) {
        rt = Py_None;
      }
      else if (!strcmp(name,"parentNode")) {
        rt = self->parentNode;
      }
      else if (!strcmp(name,"previousSibling")) {
        /*Calculate this*/
        pNode = self->parentNode;
        if (PyDocument_Check(pNode)) {
          childNodes = ((PyDocumentObject *)pNode)->childNodes;
        } else if (PyElement_Check(pNode)) {
          childNodes = ((PyElementObject *)pNode)->childNodes;
        } else if (PyDocumentFragment_Check(pNode)) {
          childNodes = ((PyDocumentFragmentObject *)pNode)->childNodes;
        } else {
          Py_INCREF(Py_None);
          return Py_None;
        }
        for (temp = 0;temp < PyList_GET_SIZE(childNodes);temp ++) {
          if ((PyNodeObject *)PyList_GET_ITEM(childNodes,temp) == self) {
            pNodeIndex = temp;
            break;
          }
        }
        if (pNodeIndex == -1) {
          /*Through error????*/
          Py_INCREF(Py_None);
          return Py_None;
        }
        if (pNodeIndex == 0) {
          /*Beginning of the list*/
          rt = Py_None;
        } else {
          rt = PyList_GET_ITEM(childNodes,pNodeIndex-1);
        }
      }
      break;
    }
    /* handle node type constants */
  case 'A': 
    {
      if (!strcmp(name, "ATTRIBUTE_NODE")) {
	rt = g_attributeNodeType;
      }
      break;
    }
  case 'C':
    {
      if (!strcmp(name, "CDATA_SECTION_NODE")) {
	rt = g_cdataSectionNodeType;
      }
      else if (!strcmp(name, "COMMENT_NODE")) {
        rt = g_commentNodeType;
      }
      break;
    }
  case 'D':
    {
      if (!strcmp(name, "DOCUMENT_NODE")) {
        rt = g_documentNodeType;
      }
      else if (!strcmp(name, "DOCUMENT_TYPE_NODE")) {
        rt = g_documentTypeNodeType;
      }
      else if (!strcmp(name, "DOCUMENT_FRAGMENT_NODE")) {
        rt = g_documentFragmentNodeType;
      }
      break;
    }
  case 'E':
    {
      if (!strcmp(name, "ELEMENT_NODE")) {
        rt = g_elementNodeType;
      }
      else if (!strcmp(name, "ENTITY_REFERENCE_NODE")) {
        rt = g_entityReferenceNodeType;
      }
      else if (!strcmp(name, "ENTITY_NODE")) {
        rt = g_entityNodeType;
      }
      break;
    }
  case 'N': 
    {
      if (!strcmp(name, "NOTATION_NODE")) {
	rt = g_notationNodeType;
      }
      break;
    }
  case 'P': 
    {
      if (!strcmp(name, "PROCESSING_INSTRUCTION_NODE")) {
	rt = g_processingInstructionNodeType;
      }
      break;
    }
  case 'T': 
    {
      if (!strcmp(name, "TEXT_NODE")) {
	rt = g_textNodeType;
      }
      break;
    }

  default:
    {
      
      if (!strcmp(name,"docIndex")) {
        rt = PyLong_FromLong(self->docIndex);
      }
      else if (!strcmp(name, "xmlBase")) {
        if (PyDocument_Check(self)) {
          rt = (PyObject *)((PyDocumentObject *)self)->refUri;
        } else if (PyElement_Check(self)) {
          rt = ((PyElementObject *)self)->xmlBase;
        } else if (PyDocumentFragment_Check(self)) {
          /* FIXME: Check that this is what we want */
          rt = Py_None;
        } else {
          pNode = self->parentNode;
          if (PyDocument_Check(pNode)) {
            rt = (PyObject *)((PyDocumentObject *)pNode)->refUri;
          } else if (PyElement_Check(pNode)) {
            rt = ((PyElementObject *)pNode)->xmlBase;
          } else if (PyDocumentFragment_Check(pNode)) {
            /* FIXME: Check that this is what we want */
            rt = Py_None;
          }
        }
      }
      break;
    }
    break;
  }
  if (rt != NULL) {
    Py_INCREF(rt);
    return rt;
  }

  return Py_FindMethod(methods, (PyObject *)self, name);
}


void *node_foreach_child(PyNodeObject *self,
                          int func(PyObject *node, void *data),
                          void *data)
{
  int ctr;
  PyObject *children;

  NODE_CHECK(self);
  /*Only docs, elements and dfs*/
  if (PyDocument_Check(self)) {
    children = ((PyDocumentObject *)self)->childNodes; 
  } else if (PyElement_Check(self)){
    children = ((PyElementObject *)self)->childNodes; 
  } else if (PyDocumentFragment_Check(self)){
    children = ((PyDocumentFragmentObject *)self)->childNodes; 
  } else {
    PyErr_SetString(g_hierarchyRequestErr, "Node has no children");
    return NULL;
  }

  for (ctr=0; ctr < PyList_GET_SIZE(children); ctr++) {
    if (!func(PyList_GET_ITEM(children, ctr), data))
      return NULL;
  }
  return self;
}


int node_to_xml(PyObject *node, PyObject *encoding)
{
  static PyObject *Print = NULL;
  PyObject *modules = PyImport_GetModuleDict();
  PyObject *TranslateCdata;

  if (!Print) {
    Print = PyDict_GetItemString(modules, "Ft.Xml.Lib.Print");
  }
  if (!Print) {
    Print = PyImport_ImportModule("Ft.Xml.Lib.Print");
  }
  TranslateCdata = PyDict_GetItemString(modules, "TranslateCdata");

  if (!encoding){
    encoding = PyUnicode_DecodeUTF8 ("UTF-8", 5, NULL);
  }

  if (PyDocument_Check(node)) {
    /*children = ((PyDocumentObject *)node)->childNodes; */
  } else if (PyElement_Check(node)){
    /* children = ((PyElementObject *)node)->childNodes; */
  } else if (PyDocumentFragment_Check(node)){
    /* children = ((PyDocumentFragmentObject *)node)->childNodes; */ 
  } else {
    PyErr_SetString(g_errorObject, "Internal error: node_to_xml");
    return 0;
  }
  return 0;
  
}


int node_test_ref_counts(PyObject *tester,PyNodeObject *node,long *childCtr,PyObject *internCtr,int base) {

  char buf[256];
  /*Inc the child ctr*/
  (*childCtr)++;

  if (PyElement_Check(node)) {
    return element_test_ref_counts(tester,(PyElementObject *)node,childCtr,internCtr,base);
  } else if (PyText_Check(node)) {
    return text_test_ref_counts(tester,(PyTextObject *)node,childCtr,internCtr,base);
  } else if (PyComment_Check(node)) {
    return comment_test_ref_counts(tester,(PyCommentObject *)node,childCtr,internCtr,base);
  } else if (PyAttr_Check(node)) {
    return attr_test_ref_counts(tester,(PyAttrObject *)node,childCtr,internCtr,base);
  } else if (PyProcessingInstruction_Check(node)) {
    return pi_test_ref_counts(tester,(PyProcessingInstructionObject *)node,childCtr,internCtr,base);
  } else {
    sprintf(buf,"Untested type: %s",node->ob_type->tp_name);
    PyObject_CallMethod(tester,"warning","s",buf);
  }

  return 1;
}
