#include "Python.h"

void SplitQName(PyObject *qualifiedName, PyObject **prefix, PyObject **localName)
{
  int i;
  Py_UNICODE colon = ':';
  int len = PyUnicode_GET_SIZE(qualifiedName);
  const Py_UNICODE *p = PyUnicode_AS_UNICODE(qualifiedName);
  
  for (i = len-1; i > -1; i--) {
    if (p[i] == colon) {
      *prefix = PyUnicode_FromUnicode(p, i);
      /* skip over the colon */
      i++;
      *localName = PyUnicode_FromUnicode((p + i), (len - i));
      return;
    }
  }

  /* No prefix */
  *prefix = Py_None;
  Py_INCREF(Py_None);
  *localName = qualifiedName;
  Py_INCREF(qualifiedName);
  return;
}

void AddInternCtr(PyObject *obj, PyObject *dict)
{
  PyObject *value;

  if (obj != Py_None) {
    value = PyDict_GetItem(dict, obj);
    if (value == NULL) {
      /*Add it  also adds a ref*/
      value = PyInt_FromLong(1);
      PyDict_SetItem(dict, obj, value);
      Py_DECREF(value);
    } else {
      value = PyInt_FromLong(PyInt_AsLong(value) + 1);
      PyDict_SetItem(dict, obj, value);
      Py_DECREF(value);
    }
  }
}

int TestRefCount(PyObject *tester,PyObject *obj,int expected, const char* name)
{
  char buf[500];
  PyObject *retval = PyObject_Repr(obj);
  if (retval == NULL) return 0;
  sprintf(buf, "%.200s equal to %.200s", name, PyString_AS_STRING(retval));
  Py_DECREF(retval);

  retval = PyObject_CallMethod(tester, "startTest", "s", buf);
  if (retval == NULL) return 0;
  Py_DECREF(retval);

  retval = PyObject_CallMethod(tester, "compare", "ll", (long)expected, obj->ob_refcnt);
  if (retval == NULL) return 0;
  Py_DECREF(retval);

  retval = PyObject_CallMethod(tester, "testDone", NULL);
  if (retval == NULL) return 0;
  Py_DECREF(retval);
  return 1;
}

